r"""
Partial permutation on `\{0, 1, ..., n-1\}` as lists.

Permutations are implemented as lists or arrays. When
the image is undefined, the number -1 is used.
"""
#*****************************************************************************
#       Copyright (C) 2019 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

from __future__ import print_function, absolute_import
from six.moves import range

from math import log

import numbers

try:
    import sage.all
except ImportError:
    from random import shuffle, randint
else:
    from sage.misc.prandom import shuffle, randint
    from sage.arith.functions import lcm

def argmin(l):
    r"""
    Return the position of the minimal element in the list ``l``.

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import argmin
        sage: argmin([3,0,1,2])
        1
        sage: argmin([-1,3,5,-2,50])
        3
    """
    if not(l):
        raise ValueError('empty list')
    imin = 0
    jmin = l[0]
    for i,j in enumerate(l):
        if j < jmin:
            jmin = j
            imin = i
    return imin

#####################################################################
# Initialization
#####################################################################

def permutation_to_perm(p):
    r"""
    Return a list on `[0, n-1]` from a permutation on `[1, n]`

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import permutation_to_perm
        sage: permutation_to_perm(PermutationGroupElement([3,1,2]))
        [2, 0, 1]
    """
    return list(map(lambda x: x-1, p.domain()))

def perm_to_permutation(l):
    r"""
    Returns a permutation on `[1, n]` from a list on `[0, n-1]`

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import perm_to_permutation
        sage: perm_to_permutation([2,1,0])
        (1,3)
    """
    try:
        # Trac #28652: Rework the constructor of PermutationGroupElement
        from sage.groups.perm_gps.constructor import PermutationGroupElement
    except ImportError:
        from sage.groups.perm_gps.permgroup_element import PermutationGroupElement
    return PermutationGroupElement(list(map(lambda x: x+1, l)))


def perm_init(data, n=None, partial=False):
    """
    Returns a permutation from ``data``.

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import perm_init
        sage: perm_init([3,2,1,0])
        [3, 2, 1, 0]
        sage: perm_init([3,2,1,0], 5)
        [3, 2, 1, 0, 4]
        sage: perm_init(([2,1],[3,4,0]))
        [3, 2, 1, 4, 0]
        sage: perm_init('(0,1)(3,2)')
        [1, 0, 3, 2]
        sage: perm_init([3,1,-1,0])
        [3, 1, -1, 0]
        sage: perm_init([[2,1],[3,4,0]])
        [3, 2, 1, 4, 0]

        sage: perm_init([0r])
        [0]

    Zerology::

        sage: perm_init([])
        []
        sage: perm_init([], 4)
        [0, 1, 2, 3]
        sage: perm_init([[]])
        []
        sage: perm_init([[]], 4)
        [0, 1, 2, 3]
    """
    if isinstance(data, (tuple,list)):
        if not data:
            return list(range(n)) if n is not None else []
        if isinstance(data[0], (tuple,list)):
            return cycles_to_list(data, n, partial)
        else:
            p = [int(x) for x in data]
            if n is not None:
                p.extend(range(len(p),n))
            return p

    if isinstance(data, str):
        c = str_to_cycles(data)
        return cycles_to_list(c, n, partial)

    raise TypeError("The input must be list, tuple or string")

def init_perm(l):
    from warnings import warn
    warn('[surface_dynamics] init_perm is deprecated use perm_init instead', Warning, 3)
    return perm_init(l)

def equalize_perms(l):
    """
    Extend the permutations in ``l`` to have the same lengths.

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import equalize_perms
        sage: li = [[0, 1], [2, 1, 0]]
        sage: equalize_perms(li)
        3
    """
    n = max(map(len, l))
    for p in l:
        p.extend(range(len(p), n))
    return n


def perm_check(l, n=None):
    r"""
    Checks that ``l`` is a partial permutation of `\{0, 1, ..., n-1\}`.

    INPUT:

    - ``n`` - integer (optional)

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import perm_check

    Good permutations::

        sage: perm_check([1, 0, 3, 2], 4)
        True
        sage: perm_check([-1], 1)
        True
        sage: perm_check([-1, 3, -1, 1], 4)
        True

        sage: perm_check([1,0,-1,-1,-1], 2)
        True

        sage: perm_check([0,-1])
        True
        sage: perm_check([-1,1])
        True

    Bad permutations::

        sage: perm_check([1, 0, 3, 2], 3)
        False
        sage: perm_check([2, 0])
        False
        sage: perm_check([1, 0, 1])
        False

        sage: perm_check([1,-1])
        False
        sage: perm_check([-1,0])
        False
    """
    if not isinstance(l, (tuple, list)):
        return False
    if n is None:
        n = len(l)
    else:
        n = int(n)

    im_seen = [False] * n
    ra_seen = [False] * n
    for i in range(n):
        if l[i] == -1:
            continue
        if not isinstance(l[i], numbers.Integral):
            raise TypeError("entries must be integers >= -1, not {}".format(type(l[i]).__name__))
        if l[i] < 0 or l[i] >= n or im_seen[l[i]]:
            return False
        ra_seen[i] = True
        im_seen[l[i]] = True
    return ra_seen == im_seen

def perm_is_one(l, n=None):
    r"""
    Test whether ``l`` is the identity on its domain.

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import perm_is_one
        sage: perm_is_one([])
        True
        sage: perm_is_one([-1])
        True
        sage: perm_is_one([0])
        True
        sage: perm_is_one([1, 0])
        False

        sage: perm_is_one([0, -1])
        True
        sage: perm_is_one([-1, 1])
        True

        sage: perm_is_one([0, 2, 1], 1)
        True
    """
    if n is None:
        n = len(l)
    for i in range(n):
        if l[i] != -1 and l[i] != i:
            return False
    return True

#####################################################################
# Conversion
#####################################################################

def cycles_to_list(t, n=None, partial=False):
    r"""
    Returns a permutation on `[0, n-1]` from a list of cycles on `[0, n-1]`

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import cycles_to_list

        sage: cycles_to_list([[1,3,5]])
        [0, 3, 2, 5, 4, 1]
        sage: cycles_to_list([[1,3,5]], partial=True)
        [-1, 3, -1, 5, -1, 1]

        sage: cycles_to_list([[1,3,5],[0,2,4],[6]])
        [2, 3, 4, 5, 0, 1, 6]

        sage: cycles_to_list([])
        []
        sage: cycles_to_list([[],[]])
        []

        sage: cycles_to_list([[1,4,2,5],[3],[6]], partial=True)
        [-1, 4, 5, 3, 2, 1, 6]

        sage: cycles_to_list([[0,5]], 3)
        Traceback (most recent call last):
        ...
        ValueError: cycle value out of range
    """
    if not any(tt for tt in t):
        return list(range(n)) if n is not None else []

    if n is None:
        n = max(map(max, t)) + 1
    if partial:
        res = [-1] * n
    else:
        res = list(range(n))

    for c in t:
        for j in range(len(c)-1):
            if not isinstance(c[j], numbers.Integral) or c[j] < 0 or c[j] >= n:
                raise ValueError("cycle values out of range")
            res[c[j]] = int(c[j+1])
        if not isinstance(c[-1], numbers.Integral) or c[-1] < 0 or c[-1] >= n:
            raise ValueError("cycle value out of range")
        res[c[-1]] = int(c[0])

    return res


def str_to_cycles(s):
    """
    Returns a list of cycles from a string.

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import str_to_cycles
        sage: str_to_cycles('(0,1)')
        [[0, 1]]
        sage: str_to_cycles('(0,1)(3,2)')
        [[0, 1], [3, 2]]

        sage: str_to_cycles('()(0,1)()(2,3)')
        [[0, 1], [2, 3]]
    """
    r = []
    for c_str in s[1:-1].split(')('):
        if not c_str:
            continue
        r.append(list(map(int, c_str.replace(' ', '').split(','))))
    return r


def perm_cycles(p, singletons=False, n=None):
    r"""
    Return the cycle decomposition of ``p``

    INPUT:

    - ``p`` -- the permutation

    - ``singletons`` -- bool (default: ``False``) - return or not the singletons

    - ``n`` - (optional) only use the first ``n`` elements of the permutation ``p``

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import perm_cycles, perm_init
        sage: perm_cycles([0,2,1])
        [[1, 2]]
        sage: perm_cycles([0,2,1],True)
        [[0], [1, 2]]

        sage: perm_cycles([2,-1,0])
        [[0, 2]]

        sage: perm_cycles([2,0,1,-1,-1], n=3)
        [[0, 2, 1]]

        sage: perm_cycles([-1,3,-1,1], singletons=True)
        [[1, 3]]

        sage: perm_cycles(perm_init('(1,3)', partial=True))
        [[1, 3]]

        sage: perm_cycles([-1, 7, -1, 5, -1, 3, -1, 1, -1, -1, -1, -1])
        [[1, 7], [3, 5]]
    """
    if n is None:
        n = len(p)
    seen = [1] * n
    res = []

    for i in range(n):
        if seen[i] and p[i] != -1:
            cycle = []
            j = i
            while seen[j]:
                seen[j] = 0
                cycle.append(j)
                j = p[j]
            if singletons or len(cycle) > 1:
                res.append(cycle)

    return res

def perm_cycle_tuples(*args, **kwds):
    print("WARNING: perm_cycle_tuples is deprecated, use perm_cycles")
    return perm_cycles(*args, **kwds)

def perm_num_cycles(p, n=None):
    r"""
    Return the number of cycles of the permutation ``p``.

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import perm_num_cycles
        sage: perm_num_cycles([1,2,3,0])
        1
        sage: perm_num_cycles([0,2,3,1])
        2
        sage: perm_num_cycles([3,2,1,0])
        2
        sage: perm_num_cycles([3,1,2,0])
        3
        sage: perm_num_cycles([0,1,2,3])
        4
    """
    if n is None:
        n = len(p)
    seen = [False] * n
    ans = 0
    for i in range(n):
        if seen[i] or p[i] == -1:
            continue
        ans += 1
        j = i
        while not seen[j]:
            seen[j] = True
            j = p[j]
    return ans

def perm_cycle_type(p, n=None):
    r"""
    Return the lengths of the cycles of the permutation ``p`` in size of
    decreasing order.

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import perm_cycle_type
        sage: perm_cycle_type([1,2,3,0])
        [4]
        sage: perm_cycle_type([0,2,3,1])
        [3, 1]
        sage: perm_cycle_type([3,2,1,0])
        [2, 2]
        sage: perm_cycle_type([3,1,2,0])
        [2, 1, 1]
        sage: perm_cycle_type([0,1,2,3])
        [1, 1, 1, 1]
    """
    if n is None:
        n = len(p)
    seen = [False] * n
    c = []
    for i in range(n):
        if seen[i] or p[i] == -1:
            continue
        k = 0
        j = i
        while not seen[j]:
            seen[j] = True
            k += 1
            j = p[j]
        c.append(k)
    c.sort(reverse=True)
    return c

def perm_dense_cycles(p, n=None):
    r"""
    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import perm_dense_cycles

        sage: perm_dense_cycles([1,2,0])
        ([0, 0, 0], [3])

        sage: perm_dense_cycles([0,2,1])
        ([0, 1, 1], [1, 2])

        sage: perm_dense_cycles([2,1,0])
        ([0, 1, 0], [2, 1])
    """
    if n is None:
        n = len(p)
    deg = []
    res = [-1] * n
    k = 0
    for i in range(n):
        if res[i] != -1:
            continue
        d = 0
        while res[i] == -1:
            res[i] = k
            i = p[i]
            d += 1
        k += 1
        deg.append(d)
    return res, deg


def perm_dense_cycles_and_angles(p, n=None):
    if n is None:
        n = len(p)

    lab = [-1] * n  # labels
    ang = [-1] * n  # angle
    deg = []          # degrees
    k = 0
    for i in range(n):
        if p[i] == -1 or lab[i] != -1:
            continue
        d = 0
        while lab[i] == -1:
            lab[i] = k
            ang[i] = d
            d += 1
            i = p[i]
        k += 1
        deg.append(d)
    return lab, ang, deg

def perm_cycle_string(p, singletons=False, n=None):
    r"""
    Returns a string representing the cycle decomposition of `p`

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import perm_cycle_string
        sage: perm_cycle_string([0,2,1])
        '(1,2)'
        sage: perm_cycle_string([0,2,1], True)
        '(0)(1,2)'
        sage: perm_cycle_string([0,1,2,-1], False, 3)
        '()'
    """
    c = perm_cycles(p, singletons, n)
    if not c:
        return '()'
    else:
        return ''.join(map(lambda x: '('+','.join(map(str, x))+')', c))

def _canonical_reg_perm(n, k):
    r"""
    Return a standard product of disjoint k-cycles on {0, 1, ..., n-1}

    TESTS::

        sage: from surface_dynamics.misc.permutation import _canonical_reg_perm
        sage: _canonical_reg_perm(6, 2)
        [1, 0, 3, 2, 5, 4]
        sage: _canonical_reg_perm(6, 3)
        [1, 2, 0, 4, 5, 3]
    """
    if not isinstance(n, numbers.Integral) or not isinstance(k, numbers.Integral):
        raise ValueError
    if n <= 0 or k <= 0 or n % k:
        raise ValueError

    p = []
    for i in range(0,n,k):
        p.extend(range(i+1, i+k))
        p.append(i)
    return p


def constellation_init(vertices, edges, faces, n=None, domain=None, check=True):
    r"""
    Each of ``vertices``, ``edges`` or ``faces can be ``None``, an
    integer or an object to initialize a (partial) permutation.

    INPUT:

    - ``vertices``, ``edges``, ``faces`` - permutations given as strings or lists

    - ``n`` - (optional) number of darts

    - ``check`` - boolean default ``True``)

    TESTS::

        sage: from surface_dynamics.misc.permutation import constellation_init

        sage: constellation_init([0,1], [1,0], [1,0])
        [[0, 1], [1, 0], [1, 0]]
        sage: constellation_init([0r,1r],[1r,0r], [1r,0r])
        [[0, 1], [1, 0], [1, 0]]

        sage: constellation_init([2,1,0], [1,2,0], None)
        [[2, 1, 0], [1, 2, 0], [0, 2, 1]]

        sage: constellation_init(3, 2, None, 6)
        [[1, 2, 0, 4, 5, 3], [1, 0, 3, 2, 5, 4], [0, 2, 5, 1, 4, 3]]

        sage: constellation_init('(0,1)', '(0,1)', '()')
        [[1, 0], [1, 0], [0, 1]]

        sage: constellation_init('(0,2,3,6)(1,4,5,7)', None, None)
        [[2, 4, 3, 6, 5, 7, 0, 1], [1, 0, 3, 2, 5, 4, 7, 6], [7, 6, 2, 0, 4, 1, 5, 3]]

    Each number (including fixed point) should be specified otherwise they are just
    ignored::

        sage: constellation_init('(1,6,5)(2,3,4)', '(1,2)(3,4)(5,6)', '(1,4,2,5)(3)(6)')
        [[-1, 6, 3, 4, 2, 1, 5], [-1, 2, 1, 4, 3, 6, 5], [-1, 4, 5, 3, 2, 1, 6]]

    TESTS::

        sage: constellation_init(None, '(0,1)(2,3)(4,5)', '(0,2,4)(1)(3,5)')
        [[5, 0, 1, 4, 3, 2], [1, 0, 3, 2, 5, 4], [2, 1, 4, 5, 0, 3]]
    """
    nones = [p is None for p in [vertices, edges, faces]]
    if sum(nones) > 1:
        if edges is None:
            edges = 2
            nones[1] = False
        else:
            raise ValueError("at least two of the three data should be provided")

    nums = [isinstance(p, numbers.Integral) for p in [vertices, edges, faces]]
    if sum(nums) == 3 and n is None:
        raise ValueError("numbers not enough to construct a constellation")

    if not (nones[0] or nums[0]):
        vertices = perm_init(vertices, n, partial=True)
        if check and not perm_check(vertices):
            raise ValueError("invalid vertex permutation")

    if not (nones[1] or nums[1]):
        edges = perm_init(edges, n, partial=True)
        if check and not perm_check(edges):
            raise ValueError("invalid edge permutation")

    if not (nones[2] or nums[2]):
        faces = perm_init(faces, n, partial=True)
        if check and not perm_check(faces):
            raise ValueError("invalid face permutation")

    P = [vertices, edges, faces]
    perms = [p for i,p in enumerate(P) if not (nones[i] or nums[i])]
    if perms:
        equalize_perms(perms)
        j = 0
        for i in range(3):
            if not (nones[i] or nums[i]):
                P[i] = perms[j]
                j += 1
        if n is None:
            n = len(perms[0])

    for i in range(3):
        if nums[i]:
            P[i] = _canonical_reg_perm(n, P[i])

    for i in range(3):
        if nones[i]:
            j = (i - 1) % 3
            k = (i + 1) % 3
            P[i] = perm_compose(perm_invert(P[j]), perm_invert(P[k]))

    if check:
        if set(P[0]) != set(P[1]) or set(P[0]) != set(P[2]):
            raise ValueError("permutations defined on different domains")
        if not perm_is_one(perm_compose(perm_compose(P[0], P[1]), P[2])):
            raise ValueError("product is not identity")

    return P

#####################################################################
# Group operations
#####################################################################

def perm_invert(l, n=None):
    r"""
    Returns the inverse of the permutation ``l``.

    TESTS::

        sage: from itertools import permutations
        sage: from surface_dynamics.misc.permutation import perm_init, perm_invert, perm_compose, perm_is_one
        sage: all(perm_is_one(perm_compose(perm_invert(p),p)) for p in permutations(range(3)))
        True
        sage: all(perm_is_one(perm_compose(p,perm_invert(p))) for p in permutations(range(3)))
        True

        sage: perm_invert([2, -1, 5, 0, -1, 3])
        [3, -1, 0, 5, -1, 2]

        sage: p = perm_init('(1,3,5)(2,9)', partial=True)
        sage: q = perm_invert(p)
        sage: q
        [-1, 5, 9, 1, -1, 3, -1, -1, -1, 2]
        sage: perm_is_one(perm_compose(p, q))
        True

        sage: perm_invert([2,1,0,-1,-1], 3)
        [2, 1, 0]
    """
    if n is None:
        n = len(l)
    res = [-1] * n
    for i in range(n):
        if l[i] != -1:
            res[l[i]] = i
    return res

def perm_invert_inplace(p, n=None):
    r"""
    Inverse in place the permutation p
    """
    if n is None:
        n = len(l)
    seen = [0] * n
    for i in range(n):
        if seen[i]:
            continue
        j = p[i]
        while not seen[i]:
            seen[i] = 1
            i, p[j] = p[j], i

def perm_compose(p1, p2):
    r"""
    Returns the product ``p1 p2``.

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import perm_init, perm_compose
        sage: perm_compose([0,2,1], [0,2,1])
        [0, 1, 2]
        sage: perm_compose([-1,2,3,1], [-1,2,1,3])
        [-1, 1, 3, 2]

        sage: p1 = perm_init('(1,3,5)', partial=True)
        sage: p2 = perm_init('(1,5)(3)', partial=True)
        sage: perm_compose(p1, p2)
        [-1, 3, -1, 1, -1, 5]
    """
    r = [-1] * len(p1)
    for i in range(len(p1)):
        if p1[i] != -1 and p1[i] < len(p2):
            r[i] = p2[p1[i]]
    return r


def perm_compose_i(p1, p2, n=None):
    r"""
    Returns the product `p1^{-1} p2^{-1}`.

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import perm_compose_i
        sage: perm_compose_i([0,1,2],[1,2,0])
        [2, 0, 1]

        sage: from surface_dynamics.misc.permutation import perm_invert, perm_compose
        sage: from itertools import permutations
        sage: for p1 in permutations(range(4)):
        ....:     for p2 in permutations(range(4)):
        ....:         assert perm_compose_i(p1, p2) == perm_compose(perm_invert(p1), perm_invert(p2))
    """
    if n is None:
        n = len(p1)

    res = [-1] * n
    for i in range(len(p1)):
        if p2[i] != -1 and p1[p2[i]] != -1:
            res[p1[p2[i]]] = i

    return res

def perm_commutator(p1, p2, n=None):
    r"""
    Return the commutator ``~p1 ~p2 p1 p2``.

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import perm_commutator, perm_compose, perm_compose_i

        sage: p1 = [1, 2, 0, 3]
        sage: p2 = [0, 3, 1, 2]
        sage: perm_commutator(p1, p2)
        [2, 3, 0, 1]
        sage: perm_compose(perm_compose_i(p1, p2), perm_compose(p1, p2))
        [2, 3, 0, 1]

        sage: p1 = [3, 2, 0, 5, 1, 4]
        sage: p2 = [5, 2, 1, 0, 4, 3]
        sage: perm_commutator(p1, p2)
        [1, 2, 5, 3, 0, 4]
        sage: perm_compose(perm_compose_i(p1, p2), perm_compose(p1, p2))
        [1, 2, 5, 3, 0, 4]

        sage: p1 = [-1, 4, -1, -1, 1, 5, 8, -1, 6]
        sage: p2 = [-1, 8, -1, -1, 4, 6, 5, -1, 1]
        sage: perm_commutator(p1, p2)
        [-1, 8, -1, -1, 5, 1, 4, -1, 6]
        sage: perm_compose(perm_compose_i(p1, p2), perm_compose(p1, p2))
        [-1, 8, -1, -1, 5, 1, 4, -1, 6]
    """
    if n is None:
        n = len(p1)
    res = [-1] * n
    for i in range(n):
        if p1[i] != -1 and p2[i] != -1 and p1[p2[i]] != -1 and p2[p1[i]] != -1:
            res[p1[p2[i]]] = p2[p1[i]]
    return res

# can we do that inplace?
def perm_conjugate(p1, p2, n=None):
    r"""
    Conjugate ``p1`` by ``p2``.

    Let call ``res`` the output of this function. If ``p1`` was
    mapping ``a`` to ``b`` then ``res`` will map ``p2[a]``
    to ``p2[b]``.

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import perm_random, perm_conjugate

        sage: p1 = perm_random(23)
        sage: p2 = perm_random(23)
        sage: res = perm_conjugate(p1, p2)
        sage: res[p2[14]] == p2[p1[14]]
        True
        sage: res[p2[19]] == p2[p1[19]]
        True
    """
    if n is None:
        n = len(p1)
    res = [-1] * n
    for i in range(n):
        res[p2[i]] = p2[p1[i]]
    return res

def perm_conjugate_inplace(p1, p2, n=None):
    r"""
    we want

    p1[p2[i]] <- p2[p1[i]]

    save tmp = p1[p2[0]]
    set  p1[p2[0]] = p2[p1[0]]

    """
    if n is None:
        n = len(p1)
    unseen = [True] * n
    for i in range(n):
        if not unseen[i]:
            continue
        unseen[i] = True

def perm_on_list_inplace(p, a, n=None, swap=None):
    r"""
    Inplace action of permutation on list-like objects.

    INPUT:

    - ``p`` - permutation

    - ``a`` - list, array

    - ``n`` - (optional) size of permutation

    - ``swap`` - (optional) a swap function

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import *
        sage: l = [0,1,2,3,4]
        sage: p = [4,2,3,0,1]
        sage: perm_on_list_inplace(p,l)
        sage: l
        [3, 4, 1, 2, 0]

    Permutation action on matrix rows::

        sage: m1 = matrix(ZZ, 5, 5, 1)
        sage: m2 = matrix(ZZ, 5, 5, 1)
        sage: m = matrix(ZZ, 5, 5, 1)
        sage: p1 = perm_init([4,1,3,2,0])
        sage: p2 = perm_init([1,0,3,4,2])
        sage: perm_on_list_inplace(p1, m1, swap=sage.matrix.matrix0.Matrix.swap_rows)
        sage: perm_on_list_inplace(p2, m2, swap=sage.matrix.matrix0.Matrix.swap_rows)
        sage: perm_on_list_inplace(perm_compose(p1, p2), m, swap=sage.matrix.matrix0.Matrix.swap_rows)
        sage: m == m2 * m1
        True
    """
    if n is None:
        n = len(p)
    seen = [False] * n
    for i in range(n):
        if seen[i]:
            continue
        seen[i] = True
        j = p[i]
        while seen[j] is False:
            if swap:
                swap(a, i, j)
            else:
                tmp = a[i]
                a[i] = a[j]
                a[j] = tmp
            seen[j] = True
            j = p[j]

################################################################
# Various permutation constructors (including randomized ones) #
################################################################

def perm_one(n):
    r"""
    The identity permutation

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import perm_one
        sage: perm_one(0)
        []
        sage: perm_one(3)
        [0, 1, 2]
    """
    return list(range(n))

def perm_random(n):
    r"""
    Return a random permutation.

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import perm_random, perm_check
        sage: perm_check(perm_random(13), 13)
        True
    """
    r = list(range(n))
    shuffle(r)
    return r

def perm_random_centralizer(p):
    r"""
    Return a random permutation in the centralizer of ``p``.

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import perm_random, perm_compose, perm_random_centralizer
        sage: p = perm_random(10)
        sage: q = perm_random_centralizer(p)
        sage: perm_compose(p, q) == perm_compose(q, p)
        True
    """
    if not p:
        return p

    cyc = perm_cycles(p)
    cyc.sort(key = lambda c: len(c))
    i = 0
    ans = [-1] * len(p)
    while i < len(cyc):
        j = i + 1
        s = len(cyc[i])
        while j < len(cyc) and len(cyc[j]) == s:
            j += 1

        # permutation of the cycles i, i+1, ..., j-1
        m = perm_random(j - i)

        for ii in range(i, j):
            jj = i + m[ii - i]
            shift = randint(0, s - 1)  # random shift
            for k in range(len(cyc[i])):
                ans[cyc[ii][k]] = cyc[jj][(k + shift) % s]

        # next round
        i = j

    return ans

def perm_random_conjugacy_class(c):
    r"""
    Return a random permutation with given conjugacy class ``c``.

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import (perm_random_conjugacy_class,
        ....:     perm_cycle_type)

        sage: p = perm_random_conjugacy_class([5,2])
        sage: perm_cycle_type(p)
        [5, 2]

        sage: p = perm_random_conjugacy_class([7,3,3,1])
        sage: perm_cycle_type(p)
        [7, 3, 3, 1]
    """
    n = sum(c)
    r = list(range(n))
    shuffle(r)
    p = [-1] * n
    i = 0
    for k in c:
        # add a k-cycle following the list r
        for j in range(i, i+k-1):
            p[r[j]] = r[j+1]
        p[r[i+k-1]] = r[i]
        i += k
    return p

#####################################################################
# Actions
#####################################################################

def perm_orbit(p, i):
    r"""
    Returns the orbit of an integer `i` under the permutation `p`

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import perm_orbit
        sage: perm_orbit([0,3,1,2],2)
        [2, 1, 3]
    """
    res = [i]
    j = p[i]
    while j != i:
        res.append(j)
        j = p[j]
    return res


def perm_on_list(p, t):
    r"""
    Action of the permutation ``p`` on the list ``t``.

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import perm_on_list
        sage: perm_on_list([2,1,3,0], [2,1,2,0])
        [3, 1, 3, 2]
    """
    return [p[i] for i in t]


def perm_on_cyclic_list(p, t):
    r"""
    Action of the permutation ``p`` on the list ``t`` up to cyclic order.

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import perm_on_cyclic_list
        sage: perm_on_cyclic_list([0,1,2], [2,1,2])
        [1, 2, 2]
        sage: perm_on_cyclic_list([0,1], [0,1,0,0,1,0,0,0,1,1])
        [0, 0, 0, 1, 1, 0, 1, 0, 0, 1]

        sage: a = [1, 0, 3, 2, 5, 4]
        sage: perm_on_cyclic_list(a, [0, 5, 3])
        [1, 4, 2]
        sage: perm_on_cyclic_list(a, [1, 4, 2])
        [0, 5, 3]

        sage: a1 = [0, 1, 4, 2, 3, 5]
        sage: a2 = [1, 5, 3, 4, 2, 0]
        sage: a3 = [2, 3, 1, 5, 0, 4]
        sage: a4 = [5, 0, 2, 3, 4, 1]
        sage: t1 = [0, 5, 1]
        sage: t2 = [2, 4, 3]
        sage: perm_on_cyclic_list(a1, t1) == perm_on_cyclic_list(a2, t1) == perm_on_cyclic_list(a4, t1) == t1
        True
        sage: perm_on_cyclic_list(a3, t1) == t2
        True
        sage: perm_on_cyclic_list(a3, t2) == t1
        True
    """
    res = r = [p[i] for i in t]
    for i in range(1, len(r)):
        rr = r[i:] + r[:i]
        if rr < res:
            res = rr
    return res


def canonical_perm(part, i=0):
    r"""
    Return the canonical permutation with the given part.

    EXAMPLES::

        sage: from surface_dynamics.flat_surfaces.separatrix_diagram import canonical_perm
        sage: canonical_perm([3,2])
        [1, 2, 0, 4, 3]
        sage: canonical_perm([2,2,2])
        [1, 0, 3, 2, 5, 4]
        sage: canonical_perm([1,1,3])
        [0, 1, 3, 4, 2]
    """
    res = []
    for p in part:
        res.extend(range(i+1,i+p))
        res.append(i)
        i += p
    return res


def canonical_perm_i(part, i=0):
    r"""
    Return the canonical permutation reversed.

    EXAMPLES::

        sage: from surface_dynamics.flat_surfaces.separatrix_diagram import canonical_perm_i
        sage: canonical_perm_i([3,2])
        [2, 0, 1, 4, 3]
        sage: canonical_perm_i([2,2,2])
        [1, 0, 3, 2, 5, 4]
        sage: canonical_perm_i([1,1,3])
        [0, 1, 4, 2, 3]
    """
    res = []
    for p in part:
        res.append(i+p-1)
        res.extend(range(i,i+p-1))
        i += p
    return res


def perm_switch(p1, p2, i, j):
    """
    Exchanges the values at positions ``i`` and ``j`` in two permutations
    ``p_1`` and ``p_2``.

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import perm_switch
        sage: a = [0,1,2]
        sage: b = [1,2,0]
        sage: perm_switch(a,b,0,1)
        sage: a;b
        [1, 0, 2]
        [2, 1, 0]
    """
    i1 = p1[i]
    j1 = p1[j]
    p1[i] = j1
    p1[j] = i1

    i2 = p2[i1]
    j2 = p2[j1]
    p2[i1] = j2
    p2[j1] = i2


def perms_transitive_components(p):
    r"""
    Return the list of transitive components of the subgroup generated by the
    permutations ``p``.

    INPUT:

    - ``p`` -- a list of permutations given as lists

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import perms_transitive_components

        sage: perms_transitive_components([[1,0,2,3],[0,1,3,2]])
        [(0, 1), (2, 3)]

        sage: perms_transitive_components([[2,3,0,1]])
        [(0, 2), (1, 3)]

        sage: perms_transitive_components([[3,1,2,0], [0,3,2,1], [0,1,3,2]])
        [(0, 1, 2, 3)]
    """
    n = len(p[0])
    seen = [-1] * n
    cc_num = 0
    for i in range(n):
        if seen[i] != -1:
            continue

        todo = [i]
        seen[i] = cc_num
        while todo:
            j = todo.pop()
            for pp in p:
                k = pp[j]
                if seen[k] == -1:
                    todo.append(k)
                    seen[k] = cc_num
        cc_num += 1

    return [tuple(i for i in range(n) if seen[i] == j) for j in range(cc_num)]

def perms_are_transitive(p):
    """
    Test whether the group generated by the permutations in ``p`` is transitive.

    INPUT:

    - ``p`` - a list of permutations of `[0, n-1]`

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import perms_are_transitive
        sage: perms_are_transitive([[0,1,2],[0,2,1]])
        False
        sage: perms_are_transitive([[0,1,2],[1,2,0]])
        True

        sage: p0 = [0,2,3,1,7,5,6,4]
        sage: p1 = [7,1,2,3,4,5,6,0]
        sage: p2 = [6,1,2,3,4,5,0,7]
        sage: p3 = [1,0,2,3,4,5,6,7]
        sage: p4 = [0,1,4,5,2,3,6,7]
        sage: perms_are_transitive([p0,p1,p2,p3,p4])
        True
        sage: perms_are_transitive([p0,p1,p2,p3])
        False
    """
    if not p:
        raise ValueError("empty list")

    n = len(p[0])
    if n <= 1:
        return True

    # compute the connected component of 0
    cc0 = [False] * n
    todo = [0]
    cc0[0] = True
    while todo:
        j = todo.pop()
        for pp in p:
            k = pp[j]
            if cc0[k] is False:
                todo.append(k)
                cc0[k] = True

    return all(j is True for j in cc0)

def perms_orbits(p, n=None):
    r"""
    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import perms_orbits

        sage: p0 = [0,2,3,1,7,5,6,4]
        sage: p1 = [7,1,2,3,4,5,6,0]
        sage: p2 = [6,1,2,3,4,5,0,7]
        sage: p3 = [1,0,2,3,4,5,6,7]

        sage: perms_orbits([p0])
        [[0], [1, 2, 3], [4, 7], [5], [6]]
        sage: perms_orbits([p0, p1])
        [[0, 7, 4], [1, 2, 3], [5], [6]]
        sage: perms_orbits([p0, p1, p2, p3])
        [[0, 7, 6, 1, 2, 3, 4], [5]]
    """
    if not p:
        raise ValueError("empty list")
    if n is None:
        n = len(p[0])
    if n == 0:
        return []
    elif n == 1:
        return [[0]]

    seen = [False] * n
    orbits = []
    for i in range(n):
        if seen[i]:
            continue
        todo = [i]
        seen[i] = True
        orbit = [i]
        while todo:
            i = todo.pop()
            for q in p:
                j = q[i]
                if seen[j] is False:
                    orbit.append(j)
                    seen[j] = True
                    todo.append(j)
        orbits.append(orbit)
    return orbits

def perms_relabel(p, m):
    """
    Relabel the list of permutations ``p`` according to ``m``.

    INPUT:

    - ``p`` - a list of permutations

    - ``m`` - the relabeling permutation

    EXAMPLES::

        sage: from surface_dynamics.misc.constellation import perms_relabel
        sage: perms_relabel([[0,1,2],[0,2,1]],[2,1,0])
        [[0, 1, 2], [1, 0, 2]]
    """
    q = [k[:] for k in p]
    for i in range(len(m)):
        for j in range(len(p)):
            q[j][m[i]] = m[p[j][i]]
    return q


def perms_canonical_labels_from(x, y, j0):
    r"""
    Return canonical labels for ``x``, ``y`` that starts at ``j0``.

    .. WARNING:

        The group generated by ``x`` and the elements of ``y`` should be
        transitive.

    INPUT:

    - ``x`` -- list - a permutation of `[0, ..., n]` as a list

    - ``y`` -- list of permutations of `[0, ..., n]` as a list of lists

    - ``j0`` -- an index in [0, ..., n]

    OUTPUT:

    mapping: a permutation that specify the new labels

    EXAMPLES::

        sage: from surface_dynamics.misc.constellation import perms_canonical_labels_from
        sage: perms_canonical_labels_from([0,1,2],[[1,2,0]],0)
        [0, 1, 2]
        sage: perms_canonical_labels_from([1,0,2],[[2,0,1]],0)
        [0, 1, 2]
        sage: perms_canonical_labels_from([1,0,2],[[2,0,1]],1)
        [1, 0, 2]
        sage: perms_canonical_labels_from([1,0,2],[[2,0,1]],2)
        [2, 1, 0]
    """
    n = len(x)

    k = 0
    mapping = [None] * n
    waiting = [[] for i in range(len(y))]

    while k < n:
        # initialize at j0
        mapping[j0] = k
        waiting[0].append(j0)
        k += 1

        # complete x cycle from j0 with labels (k, k+1, ..., k+len(cycle)-1)
        j = x[j0]
        while j != j0:
            mapping[j] = k
            waiting[0].append(j)
            k += 1
            j = x[j]

        # now run through the permutations in y to find a new element to be labeled
        l = 0
        while l < len(waiting):
            i = 0
            while i < len(waiting[l]):
                j1 = waiting[l][i]
                if mapping[y[l][j1]] is None:
                    break
                i += 1

            if i == len(waiting[l]):  # not found: go further in waiting
                if l < len(waiting)-1:
                    waiting[l+1].extend(waiting[l])
                waiting[l] = []
                l += 1
                i = 0

            else:  # found: complete cycle from new guy
                j0 = y[l][j1]
                if l < len(waiting)-1:
                    waiting[l+1].extend(waiting[l][:i+1])
                del waiting[l][:i+1]
                break

    return mapping


def perms_canonical_labels(p, e=None):
    assert(len(p) > 1)
    n = len(p[0])

    c_win = None
    m_win = list(range(n))

    x = p[0]
    y = p[1:]

    if e is None:
        e = list(range(n))

    # get canonical label from i in to_test and compare
    while e:
        i = e.pop()
        m_test = perms_canonical_labels_from(x, y, i)
        c_test = perms_relabel(p, m_test)
        if c_win is None or c_test < c_win:
            c_win = c_test
            m_win = m_test

    return c_win, m_win


#######################################################################
# Dynamical permutation groups
#######################################################################
def perm_order(p, n=None):
    r"""
    Return the multiplicative order of the permutation ``p``.

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import perm_init, perm_order
        sage: p = perm_init('(1,3)(2,4,6)(5)')
        sage: perm_order(p)
        6
    """
    return lcm(perm_cycle_type(p))

# TODO: redesign
# when adding a generator we need to check the orbit under what was already computed
class PermutationGroupOrbit(object):
    r"""
    Dynamical orbit generation.

    This is to be used when computing the automorphism group of an object (e.g.
    a ribbon graph). It allows to iterate through orbit representatives
    dynamically (i.e. group generators can be added on the fly).

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import PermutationGroupOrbit
        sage: P1 = PermutationGroupOrbit(7, ['(0,1)', '(0,1,2,3,4,5,6)'])
        sage: P2 = PermutationGroupOrbit(4, ['(0,1)', '(2,3)'])
    """
    __slots__ = ['_n', '_gens', '_S', '_s', '_seen', '_elts']

    def __init__(self, n, gens, S=None, check=True):
        if check:
            gens = [perm_init(g, n=n) for g in gens]
        if S is None:
            S = list(range(n))

        self._n = n              # underlying set is {0, 1, ..., n-1}
        self._gens = gens        # group generators
        self._S = S              # the set we are interested in
        self._s = 0              # index in the set S
        self._seen = [False] * n # the elements that we already visited (dense version)
        self._elts = []          # the elements that we already visited (sparse version)

    def __iter__(self):
        return self

    def libgap_group(self):
        from sage.libs.gap.libgap import libgap
        if not self._gens:
            return libgap.Group([libgap.PermList([])])
        gens = []
        for a in self._gens:
            gens.append(libgap.PermList([a[i]+1 for i in range(self._n)]))
        return libgap.Group(gens)

    def group_cardinality(self):
        r"""
        EXAMPLES::

            sage: from surface_dynamics.misc.permutation import PermutationGroupOrbit
            sage: PermutationGroupOrbit(7, []).group_cardinality()
            1
            sage: PermutationGroupOrbit(5, ['(0,2)(3,4)']).group_cardinality()
            2
            sage: PermutationGroupOrbit(7, ['(0,1)', '(0,1,2,3,4,5,6)']).group_cardinality()
            5040
        """
        if not self._gens:
            return 1
        elif len(self._gens) == 1:
            return perm_order(self._gens[0])
        return int(self.libgap_group().Size())

    def __repr__(self):
        n = self._n
        return "PermutationGroupOrbit({}, [{}])".format(
                n,
                ', '.join(perm_cycle_string(g, singletons=False, n=n) for g in self._gens))

    def num_gens(self):
        return len(self._gens)

    def gens(self):
        return self._gens

    def reset_iterator(self, S=None):
        r"""
        Reset the self we iterate over with.

        EXAMPLES::

            sage: from surface_dynamics.misc.permutation import PermutationGroupOrbit

            sage: O = PermutationGroupOrbit(4, [])
            sage: list(O)
            [0, 1, 2, 3]
            sage: list(O)
            []
            sage: O.reset_iterator()
            sage: list(O)
            [0, 1, 2, 3]
            sage: list(O)
            []

            sage: O.add_generator('(0,1)(2,3)')
            sage: list(O)
            []
            sage: O.reset_iterator()
            sage: list(O)
            [0, 2]
        """
        n = self._n
        if S is None:
            S = list(range(n))
        self._S = S
        self._seen = [False] * n
        del self._elts[:]
        self._s = 0

    def __next__(self):
        S = self._S        # candidates
        s = self._s        # current index in S
        if s == len(S):
            raise StopIteration

        seen = self._seen  # already visited elements (dense)
        i = S[s]
        s += 1
        while seen[i] and s < len(S):
            i = S[s]
            s += 1

        # update s and remove the orbit of i
        self._s = s
        if seen[i]:
            raise StopIteration

        elts = self._elts  # already visited elements (sparse)
        k = len(elts)
        elts.append(i)
        while k < len(elts):
            u = elts[k]
            for g in self._gens:
                v = g[u]
                if not seen[v]:
                    seen[v] = True
                    elts.append(v)
            k += 1

        return i

    next = __next__  # Python2 support

    def add_generator(self, g, check=True):
        r"""
        EXAMPLES::

            sage: from surface_dynamics.misc.permutation import PermutationGroupOrbit

            sage: O = PermutationGroupOrbit(4, [])
            sage: next(O)
            0
            sage: O.add_generator('(0,1)(2,3)')
            sage: next(O)
            2
            sage: next(O)
            Traceback (most recent call last):
            ...
            StopIteration
        """
        if check:
            g = perm_init(g, n=self._n)
        self._gens.append(g)
        elts = self._elts
        seen = self._seen

        k = 0
        K = len(elts)
        while k < K:
            u = elts[k]
            v = g[u]
            if not seen[v]:
                seen[v] = True
                elts.append(v)
            k += 1

        while k < len(elts):
            u = elts[k]
            for g in self._gens:
                v = g[u]
                if not seen[v]:
                    seen[v] = True
                    elts.append(v)
            k += 1

#################################3
# Serialization

# the first 64 characters are used for integer encoding
# the last one is used for "undefined" (corresponds to -1 in arrays)
CHARS = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ+-.'
CHARS_INV = {j:i for i,j in enumerate(CHARS)}

def uint_base64_str(n, l=None):
    r"""
    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import uint_base64_str

        sage: uint_base64_str(15)
        'f'
        sage: uint_base64_str(15, 3)
        '00f'

        sage: uint_base64_str(-1)
        '.'
        sage: uint_base64_str(-1, 3)
        '...'
    """
    n = int(n)
    if n == -1:
        if l is None:
            return CHARS[64]
        else:
            return CHARS[64] * l
    elif n < -1:
        raise ValueError("invalid negative integer")
    s = ''
    while n:
        s = CHARS[n % 64] + s
        n //= 64
    if l is not None:
        if len(s) > l:
            raise ValueError
        else:
            s = CHARS[0] * (l - len(s)) + s
    return s

def uint_from_base64_str(s):
    r"""
    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import uint_from_base64_str, uint_base64_str

        sage: uint_from_base64_str('mqb')
        91787
        sage: uint_base64_str(91787)
        'mqb'

        sage: uint_from_base64_str('00mqb')
        91787

        sage: uint_from_base64_str('...')
        -1
    """
    if not s or not isinstance(s, str):
        raise ValueError
    if s[0] == CHARS[64]:
        assert all(i == CHARS[64] for i in s)
        return -1
    n = 0
    d = 1
    for c in reversed(s):
        n += CHARS_INV[c] * d
        d *= 64
    return n

def perm_base64_str(p, n=None):
    r"""
    Make a canonical ASCII string out of ``p``.

    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import perm_base64_str, perm_from_base64_str
        sage: from array import array

        sage: perm_base64_str([])
        ''
        sage: perm_base64_str([3,1,0,2])
        '3102'
        sage: s = perm_base64_str(range(2048))
        sage: s
        '00010203...v+v-'
        sage: perm_from_base64_str(s, 2048) == list(range(2048))
        True
    """
    if n is None:
        n = len(p)
    if not n:
        return ''
    l = int(log(n, 64)) + 1 # number of digits used for each entry
    return ''.join(uint_base64_str(p[i], l) for i in range(n))

def perm_from_base64_str(s, n):
    r"""
    EXAMPLES::

        sage: from surface_dynamics.misc.permutation import perm_from_base64_str, perm_base64_str
        sage: from array import array

        sage: p = [3,0,2,1]
        sage: s = perm_base64_str(p)
        sage: perm_from_base64_str(s, 4) == p
        True

        sage: perm_base64_str([-1,-1])
        '..'
        sage: perm_from_base64_str('..', 2)
        [-1, -1]

        sage: p = list(range(3000))
        sage: shuffle(p)
        sage: perm_from_base64_str(perm_base64_str(p), 3000) == p
        True
        sage: p[18] = p[2003] = -1
        sage: perm_from_base64_str(perm_base64_str(p), 3000) == p
        True
    """
    if not n:
        if s:
            raise ValueError("invalid input")
        return []
    l = int(log(n, 64)) + 1 # number of digits used for each entry
    if len(s) != n * l:
        raise ValueError('wrong length')
    return [uint_from_base64_str(s[i:i+l]) for i in range(0,len(s),l)]



# *************************************************************************
# Copyright (C) 2015-2016 Charles Fougeron <charlesfougeron@gmail.com>
#                         Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# *************************************************************************

def mean_and_std_dev(l):
    r"""
    Return the mean and standard deviation of the floatting point numbers in
    the list ``l``.

    The implementation is very naive and should not be used for large list
    (>1000) of numbers.

    .. NOTE::

        mean and std are implemented in Sage but are quite buggy!
    """
    from math import sqrt
    m = sum(l) / len(l)
    if len(l) == 1:
        d = 0
    else:
        d = sum((x-m)**2 for x in l) / (len(l)-1)
    return m,sqrt(d)






r"""
Cardinality of Rauzy classes

The following functions implement algorithms relative to article [Del2010]_ and
[Boi2010]_ where are given formulas for the cardinality of Rauzy classes of
permutations.

  - ``c``: number of standard labeled permutations
  - ``d``: spin difference of standard labeled permutations
  - ``gamma_std``: number of standard permutations (with given profile and
    marking)
  - ``gamma_irr``: number of irreducible permutations (with given profile and
    marking)
  - ``delta_std``: spin difference for standard permutations (with given profile
    and marking)
  - ``delta_irr``: spin difference for irreducible permutations (with given
    profile and marking)

AUTHOR:

Vincent Delecroix

REFERENCES:

.. [Boi2010] Boissy 2010

.. [Del2010] Delecroix 2010

.. [Vee1982] W. Veech, "Gauss measures for transformations on the space of
   interval exchange maps", Ann. of Math., vol. 115, no. 2 (1982), pp. 201-242.

"""
#*****************************************************************************
#       Copyright (C) 2019 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

from __future__ import print_function, absolute_import
from six.moves import range

from sage.misc.cachefunc import cached_function

from sage.rings.integer import Integer

from sage.combinat.partition import Partition
from sage.arith.all import factorial, binomial

#########################
# PROFILES AND MARKINGS #
#########################

def marking_iterator(profile,left=None,standard=False):
    r"""
    Returns the marked profile associated to a partition

    EXAMPLES::

        sage: import surface_dynamics.interval_exchanges.rauzy_class_cardinality as rcc
        sage: p = Partition([3,2,2])
        sage: list(rcc.marking_iterator(p))
        [(1, 2, 0),
         (1, 2, 1),
         (1, 3, 0),
         (1, 3, 1),
         (1, 3, 2),
         (2, 2, 2),
         (2, 2, 3),
         (2, 3, 2)]
    """
    e = Partition(sorted(profile,reverse=True)).to_exp_dict()
    ek = sorted(e)

    if left is not None:
        assert(left in e)

    if left is not None: keys = [left]
    else: keys = ek

    for m in keys:
        if standard: angles = range(1,m-1)
        else: angles = range(0,m)
        for a in angles:
            yield (1,m,a)

    for m_l in keys:
        for m_r in ek:
            if m_l != m_r or e[m_l] > 1:
                yield (2,m_l,m_r)

def split(p,k,i=0):
    r"""
    Splits the i-th term of p into two parts of size k and n-k-1

    There is a symmetry split(p, k, i) = split(p, p[i]-k-1, i)

    INPUT:

    - ``p`` - a partition

    - ``k`` - an integer between 2 and p[i]

    - ``i`` - integer - the index of the element to split

    OUTPUT: a partition

    EXAMPLES::

        sage: import surface_dynamics.interval_exchanges.rauzy_class_cardinality as rcc
        sage: p = Partition([5,1])
        sage: rcc.split(p,1,0)
        [3, 1, 1]
        sage: rcc.split(p,2,0)
        [2, 2, 1]
        sage: rcc.split(p,3,0)
        [3, 1, 1]
    """
    l = list(p)
    n = l.pop(i)
    l.append(n-k-1)
    l.append(k)
    l.sort(reverse=True)
    return Partition(l)

def collapse(p,i,j):
    r"""
    Collapses the i-th term and the j-th term of a permutation

    INPUT:

    - ``p`` - a partition

    - ``i,j`` - two different indices of p

    OUTPUT:

    - a partition

    EXAMPLES::

        sage: import surface_dynamics.interval_exchanges.rauzy_class_cardinality as rcc
        sage: p = Partition([4,2,1])
        sage: rcc.collapse(p, 0, 1)
        [5, 1]
        sage: rcc.collapse(p, 0, 2)
        [4, 2]
        sage: rcc.collapse(p, 1, 2)
        [4, 2]
    """
    assert(i != j)

    l = list(p)

    n_i = l[i]
    n_j = l[j]

    l[i] = n_i+n_j-1
    del l[j]

    l.sort(reverse=True)
    return Partition(l)

def check_std_marking(p, marking):
    r"""
    Tiny internal function that checks the validity of ``marking`` on the
    partition ``p``.

    EXAMPLES::

        sage: import surface_dynamics.interval_exchanges.rauzy_class_cardinality as rcc
        sage: p = Partition([3,2,2])
        sage: rcc.check_std_marking(p, (1,3,1))
        (1, 3, 1)
        sage: rcc.check_std_marking(p, (1,3,0))
        Traceback (most recent call last):
        ...
        ValueError: marking[2] is not good
    """
    if len(marking) != 3:
        raise ValueError("marking must be a 3-tuple (t,i,j)")

    if marking[0] == 1:
        if marking[1] not in p:
            raise ValueError("marking[1] not in p")
        if marking[2] < 1 or marking[2] > marking[1]-2:
            raise ValueError("marking[2] is not good")

    elif marking[0] == 2:
        if marking[1] == marking[2]:
            if not p.to_exp()[marking[1]-1] > 1:
                raise ValueError("wrong marking type 2")
        elif marking[1] not in p or marking[2] not in p:
            raise ValueError("marking not in p")
    else:
        raise ValueError("marking[0] must be 1 or 2")

    return tuple(marking)

def check_marking(p, marking):
    r"""
    Tiny internal function that checks that ``marking`` is compatible with ``p``.

    OUTPUT:

        A 3-tuple.

    EXAMPLES::

        sage: import surface_dynamics.interval_exchanges.rauzy_class_cardinality as rcc
        sage: p = Partition([3,2,2])
        sage: rcc.check_marking(p, (1,3,1))
        (1, 3, 1)
        sage: rcc.check_marking(p, (1,3,2))
        (1, 3, 2)
        sage: rcc.check_marking(p, (1,3,3))
        Traceback (most recent call last):
        ...
        ValueError: marking[2] is not good
        sage: rcc.check_marking(p, (1,3,-1))
        Traceback (most recent call last):
        ...
        ValueError: marking[2] is not good
        sage: rcc.check_marking(p, (2,3,2))
        (2, 3, 2)
        sage: rcc.check_marking(p, (2,3,3))
        Traceback (most recent call last):
        ...
        ValueError: wrong marking type 2
    """
    if len(marking) != 3:
        raise ValueError("marking must be a 3-tuple (t,i,j)")

    if marking[0] == 1:
        if marking[1] not in p:
            raise ValueError("marking[1] not in p")
        if marking[2] < 0 or marking[2] > marking[1]-1:
            raise ValueError("marking[2] is not good")

    elif marking[0] == 2:
        if marking[1] == marking[2]:
            if not p.to_exp()[marking[1]-1] > 1:
                raise ValueError("wrong marking type 2")
        elif marking[1] not in p or marking[2] not in p:
            raise ValueError("marking not in p")
    else:
        raise ValueError("marking[0] must be 1 or 2")

    return tuple(marking)

def bidecompositions(p):
    r"""
    Iterator through the pair of partitions ``(q1,q2)`` such that the union of
    the parts of ``q1`` and ``q2`` equal ``p``.

    EXAMPLES::

        sage: import surface_dynamics.interval_exchanges.rauzy_class_cardinality as rcc
        sage: list(rcc.bidecompositions(Partition([3,1])))
        [([], [3, 1]), ([3], [1]), ([1], [3]), ([3, 1], [])]
        sage: list(rcc.bidecompositions(Partition([2,1,1])))
        [([], [2, 1, 1]),
         ([2], [1, 1]),
         ([1], [2, 1]),
         ([2, 1], [1]),
         ([1, 1], [2]),
         ([2, 1, 1], [])]
    """
    from itertools import product

    exp = p.to_exp()
    for i in product(*tuple(range(i+1) for i in exp)):
        p1 = Partition(exp=i)
        p2 = Partition(exp=[exp[j]-i[j] for j in range(len(exp))])
        yield p1,p2

#########################
# STANDARD PERMUTATIONS #
#########################

# number of permutations

@cached_function
def _c_rec(p):
    r"""
    Recurrence function that is called by :func:`c`.
    """
    if p[0] == 1: return factorial(len(p)-1)

    return (
         sum(_c_rec(split(p,k)) for k in range(1,p[0]-1)) +
         sum(p[i]*_c_rec(collapse(p,0,i)) for i in range(1,len(p))))

def c(p):
    r"""
    Number of labeled standard permutations with given profile

    There is an explicit formula for this number

    .. MATH::

        c(p) = \frac{2 (n-1)!}{n+1} \left( \sum_{q \subset (p_2,p_3,\ldots,p_k)} (-1)^{s(q)-l(q)} \binom{n}{s(q)}^{-1} \right).

    Though, for huge partition `p` this is not very useful. This function
    implements an induction formula to compute `c(p)`.

    EXAMPLES::

        sage: import surface_dynamics.interval_exchanges.rauzy_class_cardinality as rcc

    Partition of length 1::

        sage: n = 7
        sage: rcc.c([n]) == 2 * factorial(n-1) / (n+1)
        True
        sage: all(rcc.c([n]) == 2 * factorial(n-1) / (n+1) for n in range(11,18,2))
        True

    Partitions of length 2 with two odd numbers::

        sage: p = [5,3]
        sage: n = sum(p)
        sage: b = binomial(n,p[0])
        sage: rcc.c(p) == 2 * factorial(n-1) * (1 + 1 / binomial(n,p[0])) / (n+1)
        True

        sage: p = [13,5]
        sage: n = sum(p)
        sage: b = binomial(n,p[0])
        sage: rcc.c(p) == 2 * factorial(n-1) * (1 + 1 / binomial(n,p[0])) / (n+1)
        True

    Partitions of length 2 with even numbers::

        sage: p = [4,4]
        sage: n = sum(p)
        sage: b = binomial(n,p[0])
        sage: rcc.c(p) == 2 * factorial(n-1) * (1 - 1 / binomial(n,p[0])) / (n+1)
        True

        sage: p = [10,2]
        sage: n = sum(p)
        sage: b = binomial(n,p[0])
        sage: rcc.c(p) == 2 * factorial(n-1) * (1 - 1 / binomial(n,p[0])) / (n+1)
        True

    Add marked points to an integer partition::

        sage: p = [3,2,2]
        sage: n = sum(p)
        sage: all(rcc.c(p + [1]*k) == factorial(n+k-1) / factorial(n-1) * rcc.c(p) for k in range(1,6))
        True
    """
    return _c_rec(Partition(p))


def gamma_std(profile, marking=None):
    r"""
    Return the number of standard permutations of given profile

    INPUT:

    - ``profile`` - an integer partition such that the its sum plus its length
      is congruent to 0 modulo 2

    - ``marking`` - either None, an element of the profile or a 3-tuple ``(1, n1, a)`` or ``(2, n1, n2)``

    EXAMPLES::

        sage: import surface_dynamics.interval_exchanges.rauzy_class_cardinality as rcc

    A ValueError is raised if the partition does not satisfy the requirement::

        sage: rcc.gamma_std([5,2])
        Traceback (most recent call last):
        ...
        ValueError: the sum of the profile (=[5, 2]) plus its length must be congruent to 0 modulo 2

    The Rauzy classes associated to connected strata in genus 3::

        sage: from surface_dynamics import AbelianStratum
        sage: cc = AbelianStratum(1,1,1,1).unique_component()
        sage: d = cc.rauzy_diagram()
        sage: d
        Rauzy diagram with 1255 permutations
        sage: sum(1 for _ in filter(lambda x: x.is_standard(), d)) == rcc.gamma_std([2,2,2,2])
        True

        sage: cc = AbelianStratum(2,1,1).unique_component()
        sage: d = cc.rauzy_diagram()
        sage: d
        Rauzy diagram with 2177 permutations
        sage: sum(1 for _ in filter(lambda x: x.is_standard(), d)) == rcc.gamma_std([3,2,2])
        True

        sage: cc = AbelianStratum(3,1).unique_component()
        sage: d = cc.rauzy_diagram()
        sage: d
        Rauzy diagram with 770 permutations
        sage: sum(1 for _ in filter(lambda x: x.is_standard(), d)) == rcc.gamma_std([4,2])
        True

    The non connected strata in genus 3::

        sage: cc_odd = AbelianStratum(2,2).odd_component()
        sage: cc_hyp = AbelianStratum(2,2).hyperelliptic_component()
        sage: d_odd = cc_odd.rauzy_diagram()
        sage: d_hyp = cc_hyp.rauzy_diagram()
        sage: d_odd
        Rauzy diagram with 294 permutations
        sage: d_hyp
        Rauzy diagram with 63 permutations
        sage: n_odd = sum(1 for _ in filter(lambda x: x.is_standard(), d_odd))
        sage: n_hyp = sum(1 for _ in filter(lambda x: x.is_standard(), d_hyp))
        sage: n_odd + n_hyp == rcc.gamma_std([3,3])
        True

        sage: cc_odd = AbelianStratum(4).odd_component()
        sage: cc_hyp = AbelianStratum(4).hyperelliptic_component()
        sage: d_odd = cc_odd.rauzy_diagram()
        sage: d_hyp = cc_hyp.rauzy_diagram()
        sage: d_odd
        Rauzy diagram with 134 permutations
        sage: d_hyp
        Rauzy diagram with 31 permutations
        sage: n_odd = sum(1 for _ in filter(lambda x: x.is_standard(), d_odd))
        sage: n_hyp = sum(1 for _ in filter(lambda x: x.is_standard(), d_hyp))
        sage: n_odd + n_hyp == rcc.gamma_std([5])
        True
    """
    p = Partition(sorted(profile,reverse=True))
    if (sum(p) + len(p)) % 2 != 0:
        raise ValueError("the sum of the profile (=%s) plus its length must be congruent to 0 modulo 2" % p)

    if len(p) == 0 and marking==(1,0,0):
        return 1

    if marking is None:
        return sum(gamma_std(p,m) for m in marking_iterator(p,left=None,standard=True))

    elif isinstance(marking, (int,Integer)):
        return sum(gamma_std(p,m) for m in marking_iterator(p,left=marking,standard=True))

    marking = check_std_marking(p, marking)
    l = list(p)

    if marking[0] == 1: # marking of type 1
        i = l.index(marking[1])
        del l[i]
        return _c_rec(split(p,marking[2],i)) / Partition(l).centralizer_size()

    else: # marking of type 2
        i0 = l.index(marking[1])
        if marking[2] <= marking[1]:
            i1 = i0 + 1 + l[i0+1:].index(marking[2])
        else:
            i1 = l.index(marking[2])
            i0,i1 = i1,i0
        del l[i1]
        del l[i0]

        return _c_rec(collapse(p,i0,i1)) / Partition(l).centralizer_size()

number_of_standard_permutations = gamma_std

# spin difference

def d(p):
    r"""
    Difference between the number of odd spin parity and even spin parity
    standard labeled permutations with given profile

    There is an explicit formula

    .. MATH::

        d(p) = \frac{(n-1)!}{2^{(n-k)/2}}

    where `n` is the sum of the partition `p` and `k` is its length.

    EXAMPLES::

        sage: import surface_dynamics.interval_exchanges.rauzy_class_cardinality as rcc
        sage: p = [3,3,1]
        sage: rcc.d([3,3,1]) == factorial(6) / 2**2
        True
        sage: rcc.d([13]) == factorial(12) / 2**6
        True

     Adding marked points::

        sage: p = [5,3,3]
        sage: n = sum(p)
        sage: all(rcc.d(p + [1]*k) == factorial(n+k-1) / factorial(n-1) * rcc.d(p) for k in range(1,6))
        True
    """
    n = sum(p)
    k = len(p)
    return factorial(n-1) / 2**((n-k)/2)


def delta_std(profile, marking=None):
    r"""
    Return the difference odd-even in the given stratum

    INPUT:

    - ``p`` - partition with odd terms

    - ``marking`` - a 3-tuple ``(1, n1, a)`` or ``(2, n1, n2)``

    EXAMPLES::

        sage: import surface_dynamics.interval_exchanges.rauzy_class_cardinality as rcc


    A ValueError is raised if the partition does not fullfill the requirement::

        sage: rcc.delta_std([5,2])
        Traceback (most recent call last):
        ...
        ValueError: the profile (=[5, 2]) must contain only odd numbers

    Non connected strata in genus 3 has two connected components distinguished
    by their spin parity::

        sage: from surface_dynamics import AbelianStratum
        sage: cc_odd = AbelianStratum(2,2).odd_component()
        sage: cc_hyp = AbelianStratum(2,2).hyperelliptic_component()
        sage: d_odd = cc_odd.rauzy_diagram()
        sage: d_hyp = cc_hyp.rauzy_diagram()
        sage: d_odd
        Rauzy diagram with 294 permutations
        sage: d_hyp
        Rauzy diagram with 63 permutations
        sage: n_odd = sum(1 for _ in filter(lambda x: x.is_standard(), d_odd))
        sage: n_hyp = sum(1 for _ in filter(lambda x: x.is_standard(), d_hyp))
        sage: n_odd - n_hyp == rcc.delta_std([3,3])
        True

        sage: cc_odd = AbelianStratum(4).odd_component()
        sage: cc_hyp = AbelianStratum(4).hyperelliptic_component()
        sage: d_odd = cc_odd.rauzy_diagram()
        sage: d_hyp = cc_hyp.rauzy_diagram()
        sage: d_odd
        Rauzy diagram with 134 permutations
        sage: d_hyp
        Rauzy diagram with 31 permutations
        sage: n_odd = sum(1 for _ in filter(lambda x: x.is_standard(), d_odd))
        sage: n_hyp = sum(1 for _ in filter(lambda x: x.is_standard(), d_hyp))
        sage: n_odd - n_hyp == rcc.delta_std([5])
        True
    """
    p = Partition(sorted(profile,reverse=True))

    if any(not x%2 for x in p):
        raise ValueError("the profile (=%s) must contain only odd numbers" % p)

    if marking is None:
        return sum(delta_std(p,m) for m in marking_iterator(p,left=None,standard=True))

    elif isinstance(marking, (int,Integer)):
        return sum(delta_std(p,m) for m in marking_iterator(p,left=marking,standard=True))

    marking = check_std_marking(p, marking)
    l = list(p)

    if marking[0] == 1: # marking of type 1
        if marking[2]%2 == 0: return 0
        i = l.index(marking[1])
        del l[i]
        return d(split(p,marking[2],i)) / Partition(l).centralizer_size()

    else: # marking of type 2
        i0 = l.index(marking[1])
        if marking[2] <= marking[1]:
            i1 = i0 + 1 + l[i0+1:].index(marking[2])
        else:
            i1 = l.index(marking[2])
            i0,i1 = i1,i0
        del l[i1]
        del l[i0]
        return d(collapse(p,i0,i1)) / Partition(l).centralizer_size()

spin_difference_for_standard_permutations = delta_std

#####################################
# FROM STANDARD PERMUTATIONS TO ALL #
#####################################

@cached_function
def _gamma_irr_rec(p, marking):
    r"""
    Internal recursive function called by :func:`gamma_irr`
    """
    if len(p) == 0:
        return 1

    if marking[0] == 1:
        m = marking[1]
        a = marking[2]
        i = p.index(m)
        pp = Partition(p._list[:i]+p._list[i+1:]) # the partition p'

        N = gamma_std(pp._list + [m+2],(1,m+2,m-a))

        for m1 in range(1,m-1):
            m2 = m-m1-1
            for a1 in range(max(0,a-m2),min(a,m1)):
                a2 = a - a1 - 1
                for p1,p2 in bidecompositions(pp):
                    l1 = sorted([m1]+p1._list,reverse=True)
                    l2 = sorted([m2+2]+p2._list,reverse=True)
                    if (sum(l1)+len(l1)) % 2 == 0 and (sum(l2)+len(l2)) % 2 == 0:
                        N -= (_gamma_irr_rec(Partition(l1), (1,m1,a1)) *
                              gamma_std(Partition(l2),(1,m2+2,m2-a2)))
        return N


    elif marking[0] == 2:
        m1 = marking[1]
        m2 = marking[2]
        i1 = p.index(m1)
        i2 = p.index(m2)
        if m1 == m2: i2 += 1
        if i2 < i1: i1,i2 = i2,i1
        pp = Partition(p._list[:i1] + p._list[i1+1:i2] + p._list[i2+1:])

        N = gamma_std(pp._list + [m1+1,m2+1],(2,m1+1,m2+1))

        for p1,p2 in bidecompositions(pp):
            for k1 in range(1,m1): # remove (m'_1|.) (m''_1 o m_2)
                k2 = m1-k1-1
                l1 = sorted(p1._list+[k1],reverse=True)
                l2 = sorted(p2._list+[k2+1,m2+1],reverse=True)
                if (sum(l1)+len(l1)) %2 == 0 and (sum(l2)+len(l2)) %2 == 0:
                    for a in range(k1): # a is an angle
                        N -= (_gamma_irr_rec(Partition(l1), (1,k1,a))*
                              gamma_std(Partition(l2),(2,k2+1,m2+1)))

            for k1 in range(1,m2): # remove (m_1 o m'_2) (m''_2|.)
                k2 = m2-k1-1
                l1 = sorted(p1._list+[m1,k1],reverse=True)
                l2 = sorted(p2._list+[k2+2],reverse=True)
                if (sum(l1)+len(l1)) %2 == 0 and (sum(l2)+len(l2)) %2 == 0:
                    for a in range(1,k2+1): # a is an angle for standard perm
                        N -= (_gamma_irr_rec(Partition(l1), (2,m1,k1)) *
                              gamma_std(Partition(l2),(1,k2+2,a)))

        for m in pp.to_exp_dict(): # remove (m_1, k_1) (k_2, m_2) for k1+k2+1 an other zero
            q = pp._list[:]
            del q[q.index(m)]
            for p1,p2 in bidecompositions(Partition(q)):
                for k1 in range(1,m):
                    k2 = m-k1-1
                    l1 = sorted(p1._list+[m1,k1],reverse=True)
                    l2 = sorted(p2._list+[k2+1,m2+1],reverse=True)
                    if (sum(l1)+len(l1))%2 == 0 and (sum(l2)+len(l2))%2 == 0:
                        N -= (_gamma_irr_rec(Partition(l1), (2,m1,k1)) *
                              gamma_std(Partition(l2),(2,k2+1,m2+1)))

        return N

    else:
        raise ValueError("marking must be a 3-tuple of the form (1,m,a) or (2,m1,m2)")

def gamma_irr(profile=None, marking=None):
    r"""
    Number of permutations for the given profile and marking

    INPUT:

    - ``profile`` - an integer partition such that its sum plus its length is
      congruent to 0 modulo 2

    - ``markings`` - None, an element of the profile or a 3-tuple

    EXAMPLES::

        sage: import surface_dynamics.interval_exchanges.rauzy_class_cardinality as rcc

    The connected strata in genus 3::

        sage: from surface_dynamics import AbelianStratum

        sage: c = AbelianStratum(1,1,1,1).unique_component()
        sage: c.rauzy_diagram()
        Rauzy diagram with 1255 permutations
        sage: rcc.gamma_irr([2,2,2,2])
        1255

        sage: c = AbelianStratum(2,1,1).unique_component()
        sage: c.rauzy_diagram()
        Rauzy diagram with 2177 permutations
        sage: rcc.gamma_irr([3,2,2])
        2177

        sage: c = AbelianStratum(3,1).unique_component()
        sage: c.rauzy_diagram()
        Rauzy diagram with 770 permutations
        sage: rcc.gamma_irr([4,2])
        770

    The non connecte strata in genus 3::

        sage: c_odd = AbelianStratum(2,2).odd_component()
        sage: c_hyp = AbelianStratum(2,2).hyperelliptic_component()
        sage: c_odd.rauzy_diagram()
        Rauzy diagram with 294 permutations
        sage: c_hyp.rauzy_diagram()
        Rauzy diagram with 63 permutations
        sage: rcc.gamma_irr([3,3]) == 294 + 63
        True

        sage: c_odd = AbelianStratum(4).odd_component()
        sage: c_hyp = AbelianStratum(4).hyperelliptic_component()
        sage: c_odd.rauzy_diagram()
        Rauzy diagram with 134 permutations
        sage: c_hyp.rauzy_diagram()
        Rauzy diagram with 31 permutations
        sage: rcc.gamma_irr([5]) == 134 + 31
        True
    """
    p = Partition(sorted(profile,reverse=True))

    if marking is None:
        return sum(gamma_irr(profile,m) for m in marking_iterator(profile,left=None))

    elif isinstance(marking, (int,Integer)):
        return sum(gamma_irr(profile,m) for m in marking_iterator(profile,left=marking))

    marking = check_marking(p, marking)
    return _gamma_irr_rec(p, marking)

number_of_irreducible_permutations = gamma_irr

@cached_function
def _delta_irr_rec(p, marking):
    r"""
    Internal recursive function called by :func:`delta_irr`.
    """
    if len(p) == 0:
        return 0

    if marking[0] == 1:
        m = marking[1]
        a = marking[2]
        i = p.index(m)
        pp = Partition(p._list[:i]+p._list[i+1:]) # the partition p'

        N = (-1)**a* delta_std(
                pp._list + [m+2],
                (1,m+2,m-a))

        for m1 in range(1,m-1,2):
            m2 = m-m1-1
            for a1 in range(max(0,a-m2),min(a,m1)):
                a2 = a - a1 - 1
                for p1,p2 in bidecompositions(pp):
                    l1 = sorted([m1]+p1._list,reverse=True)
                    l2 = sorted([m2+2]+p2._list,reverse=True)
                    N += (-1)**a2*(_delta_irr_rec(Partition(l1),(1,m1,a1)) *
                        spin_difference_for_standard_permutations(Partition(l2),(1,m2+2,m2-a2)))
        return N

    elif marking[0] == 2:
        m1 = marking[1]
        m2 = marking[2]
        i1 = p.index(m1)
        i2 = p.index(m2)
        if m1 == m2: i2 += 1
        if i2 < i1: i1,i2 = i2,i1
        pp = Partition(p._list[:i1] + p._list[i1+1:i2] + p._list[i2+1:])

        N = d(Partition(sorted(pp._list+[m1+m2+1],reverse=True))) /  pp.centralizer_size()
        # nb of standard permutations that corrresponds to extension of good
        # guys

        for p1,p2 in bidecompositions(Partition(pp)):
            for k1 in range(1,m1,2): # remove (k1|.) (k2 o m_2)
                k2 = m1-k1-1
                q1 = Partition(sorted(p1._list+[k1],reverse=True))
                q2 = Partition(sorted(p2._list+[k2+m2+1],reverse=True))
                for a in range(k1): # a is a angle
                    N += _delta_irr_rec(q1, (1,k1,a)) * d(q2) / p2.centralizer_size()

            for k1 in range(1,m2,2): # remove (m_1 o k1) (k2|.)
                k2 = m2-k1-1
                l1 = sorted(p1._list+[m1,k1],reverse=True)
                l2 = sorted(p2._list+[k2+2],reverse=True)
                for a in range(1,k2+1): # a is an angle for standard perm
                    N += (_delta_irr_rec(Partition(l1), (2,m1,k1)) *
                        spin_difference_for_standard_permutations(Partition(l2), (1,k2+2,a)))

        for m in pp.to_exp_dict(): # remove (m_1 o k_1) (k_2 o m_2) for k1+k2+1 an other zero
            q = pp._list[:]
            del q[q.index(m)]
            for p1,p2 in bidecompositions(Partition(q)):
                for k1 in range(1,m,2):
                    k2 = m-k1-1
                    q1 = Partition(sorted(p1._list+[m1,k1],reverse=True))
                    q2 = Partition(sorted(p2._list+[k2+m2+1],reverse=True))
                    N += _delta_irr_rec(q1, (2,m1,k1)) * d(q2) / p2.centralizer_size()

        return N

def delta_irr(profile, marking=None):
    r"""
    Spin difference for the given profile and marking

    EXAMPLES::

        sage: import surface_dynamics.interval_exchanges.rauzy_class_cardinality as rcc

    The non connecte strata in genus 3::

        sage: from surface_dynamics import AbelianStratum

        sage: c_odd = AbelianStratum(2,2).odd_component()
        sage: c_hyp = AbelianStratum(2,2).hyperelliptic_component()
        sage: c_odd.rauzy_diagram()
        Rauzy diagram with 294 permutations
        sage: c_hyp.rauzy_diagram()
        Rauzy diagram with 63 permutations
        sage: rcc.delta_irr([3,3]) == 294 - 63
        True

        sage: c_odd = AbelianStratum(4).odd_component()
        sage: c_hyp = AbelianStratum(4).hyperelliptic_component()
        sage: c_odd.rauzy_diagram()
        Rauzy diagram with 134 permutations
        sage: c_hyp.rauzy_diagram()
        Rauzy diagram with 31 permutations
        sage: rcc.delta_irr([5]) == 134 - 31
        True

    A non connected strata in genus 4::

        sage: import surface_dynamics.interval_exchanges.rauzy_class_cardinality as rdc
        sage: a = AbelianStratum(6)
        sage: c_hyp = a.hyperelliptic_component()
        sage: c_odd = a.odd_component()
        sage: c_even = a.even_component()
        sage: c_hyp.rauzy_diagram()
        Rauzy diagram with 127 permutations
        sage: c_hyp.rauzy_class_cardinality()
        127
        sage: c_odd.rauzy_diagram()
        Rauzy diagram with 5209 permutations
        sage: c_odd.rauzy_class_cardinality()
        5209
        sage: c_even = a.even_component()
        sage: c_even.rauzy_diagram()
        Rauzy diagram with 2327 permutations
        sage: c_even.rauzy_class_cardinality()
        2327
        sage: 5209 - 2327 - 127
        2755
        sage: rdc.delta_irr([7])
        2755

    An example with a very big Rauzy class::

        sage: c = AbelianStratum(6,6).odd_component()
        sage: c.rauzy_class_cardinality()
        11609364656
    """
    p = Partition(sorted(profile,reverse=True))

    if marking is None:
        return sum(delta_irr(profile,m) for m in marking_iterator(profile,left=None))

    elif isinstance(marking, (int,Integer)):
        return sum(delta_irr(profile,m) for m in marking_iterator(profile,left=marking))

    marking = check_marking(p, marking)
    return _delta_irr_rec(p, marking)

spin_difference_for_irreducible_permutations = delta_irr




#*****************************************************************************
#       Copyright (C) 2019 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

from __future__ import absolute_import

from . import constructors as iet

del absolute_import



#*****************************************************************************
#       Copyright (C) 2019 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

from libc.stdint cimport uint64_t

cdef extern from "int_iet.h":
    ctypedef struct label:
        pass
    ctypedef struct int_iet:
        pass
    ctypedef int_iet int_iet_t[1]

    ctypedef struct li_vector_iterator:
        uint64_t * x
    ctypedef li_vector_iterator li_vector_iterator_t[1]

    # memory allocation
    void int_iet_init(int_iet_t t, unsigned int n)
    void int_iet_clear(int_iet_t t)

    # safety check
    int  int_iet_check(int_iet_t t)

    # set data
    void int_iet_set_labels_and_twin(int_iet_t t, int * labels, int * twin, int k)
    void int_iet_set_lengths(int_iet_t t, uint64_t * lengths)
    void int_iet_set_random_lengths(int_iet_t t, uint64_t L)

    # output
    void int_iet_print(int_iet_t t)

    # number of cylinders
    int int_iet_num_cylinders(uint64_t * widths, uint64_t * heights, int_iet_t t)

    # iterator through vectors of given sum n and length k
    int int_vector_first(uint64_t * x, int n, int k)
    int int_vector_next(uint64_t * x, int n, int k)

    # iteration through integer vectors of given sum and length
    void int_li_vector_init(li_vector_iterator_t t, uint64_t n, int kfree, int ktop, int kbot)
    void int_li_vector_clear(li_vector_iterator_t t);
    int int_li_vector_prefirst(li_vector_iterator_t t)
    int int_li_vector_first_or_next(li_vector_iterator_t t)
    void int_li_vector_info(li_vector_iterator_t t)

cdef void get_stat(object, bint, int_iet_t, uint64_t *, uint64_t *, int)



r"""
Strata of differential on Riemann surfaces

This file gather common code used in
:mod:`~surface_dynamics.flat_surfaces.abelian_strata` and
:mod:`~surface_dynamics.flat_surfaces.quadratic_strata`.
"""
#*****************************************************************************
#       Copyright (C) 2009-2019 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

from __future__ import print_function, absolute_import, division
from six.moves import range, map, filter, zip

from functools import total_ordering

from sage.structure.unique_representation import UniqueRepresentation
from sage.structure.sage_object import SageObject
from sage.structure.parent import Parent

def list_to_exp_list(l):
    r"""
    Convert list into exponential notation.

    EXAMPLES::

        sage: from surface_dynamics import *

        sage: from surface_dynamics.flat_surfaces.strata import list_to_exp_list
        sage: l = [0,0,2,2,3,2,0,0,0]
        sage: list_to_exp_list(l)
        [(0, 2), (2, 2), (3, 1), (2, 1), (0, 3)]
    """
    d = []
    i = 0
    while i < len(l):
        j = i
        while j < len(l) and l[j] == l[i]:
            j += 1
        d.append((l[i],j-i))
        i = j
    return d

#
# Stratum, stratum component
#

class Stratum(UniqueRepresentation, SageObject):
    r"""
    Generic class for stratum of flat surfaces.

    Assumes there are

    - a method ``.zeros()`` which returns the list of all zeros

    - a method ``.nb_zeros()`` which returns the number of zeros with an option
      fake_zeros which could be true or false

    - a method ``.nb_fake_zeros()`` which returns the number of fake zeros (or
      marked points)

    - a method ``.dimension()`` which returns the dimension of the stratum

    - an attribute ``._cc`` which is a list of classes associated to the
      connected components of self

    There may be

    - an attribute ``._name`` which corresponds to the begining of the string
      representation (default is the empty string)

    - an attribute ``._latex_name`` which corresponds to the begining of the latex
      string representation (uses ``_name`` by default)

    TESTS::

            sage: from surface_dynamics import *

            sage: A = AbelianStratum(2,2)
            sage: B = AbelianStratum(1,1)
            sage: hash(A) == hash(B)
            False
    """
    _name = ''
    _latex_name = ''

    #
    # String representation
    #

    def _flat_zero_str(self):
        r"""
        String representation of the zeros.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: a = AbelianStratum({2:3})
            sage: a._flat_zero_str()
            '2, 2, 2'
        """
        return ', '.join(map(str,self.zeros()))

    def _exp_zero_str(self):
        r"""
        String representation with exponential notation

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: a = AbelianStratum(2,2,2)
            sage: a._exp_zero_str()
            '2^3'
        """
        return ', '.join('%d^%d' %(i,e) if e != 1 else '%d' %i for (i,e) in list_to_exp_list(self.zeros()))

    # this attribute can be switched between _flat_zero_str and _exp_zero_str
    _zero_str = _exp_zero_str

    def _repr_(self):
        """
        TESTS::

            sage: from surface_dynamics import *

            sage: repr(AbelianStratum(1,1))       # indirect doctest
            'H_2(1^2)'
            sage: repr(QuadraticStratum(1,1,1,1)) # indirect doctest
            'Q_2(1^4)'
        """
        return self._name + "_" + str(self.genus()) + "(" + self._zero_str() + ")"

    def _latex_(self):
        r"""
        Latex string representation

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: AbelianStratum(0)._latex_()
            '\\mathcal{H}_1(0)'
            sage: QuadraticStratum({-1:4})._latex_()
            '\\mathcal{Q}_0(-1^4)'
        """
        return self._latex_name + '_' + str(self.genus()) + "(" + self._zero_str() + ")"

    #
    # Equality and comparisons
    #

    def __lt__(self, other):
        r"""
        Comparison

        ALGORITHM:

        First compare the class, then the dimension and then the list of zeros.

        TESTS::

            sage: from surface_dynamics import *

            sage: AbelianStratum(0) == AbelianStratum(0)
            True
            sage: QuadraticStratum(5,-1) == QuadraticStratum(5,-1)
            True
            sage: QuadraticStratum(5,-1) != QuadraticStratum(5,-1)
            False

            sage: AbelianStratum(12) == QuadraticStratum(12)
            False
            sage: QuadraticStratum(12,0,0) == QuadraticStratum(12,0)
            False

            sage: AbelianStratum(2,0) == AbelianStratum(2)
            False
            sage: AbelianStratum(2,0) != AbelianStratum(2)
            True

            sage: AbelianStratum(1,1) < AbelianStratum(1,1,0)
            True
            sage: AbelianStratum(1,1,0) < AbelianStratum(1,1)
            False
            sage: AbelianStratum(1,1,0) < AbelianStratum(1,1,0,0)
            True
            sage: AbelianStratum(2) < AbelianStratum(1,1)
            True
            sage: AbelianStratum(4,0) > AbelianStratum(1,1,1,1)
            False
            sage: AbelianStratum(4,0,0,0) > AbelianStratum(1,1,1,1)
            True

        ::

            sage: QuadraticStratum(2,2) < QuadraticStratum(2,2,0)
            True
            sage: QuadraticStratum(2,2,0) < QuadraticStratum(2,2)
            False
            sage: QuadraticStratum(2,2,0) < QuadraticStratum(2,2,0,0)
            True
            sage: QuadraticStratum(4) < QuadraticStratum(2,2)
            True
            sage: QuadraticStratum(4,0) > QuadraticStratum(1,1,1,1)
            False

            sage: Q1 = QuadraticStratum(4,0,0,0)
            sage: Q2 = QuadraticStratum(1,1,1,1)
            sage: Q1 > Q2
            True
            sage: Q1 >= Q2
            True
            sage: Q1 < Q2
            False
            sage: Q1 <= Q2
            False
        """
        if not isinstance(self, Stratum) or not isinstance(other, Stratum):
            raise TypeError

        if type(self) is type(other):
            # compare the dimension
            if self.dimension() < other.dimension():
                return True
            elif self.dimension() > other.dimension():
                return False

            # compare the list of zeros
            if self.zeros() < other.zeros():
                return True
            elif self.zeros() > other.zeros():
                return False

            # equality
            return False

        else:
            sname = type(self).__name__
            oname = type(other).__name__
            if sname < oname:
                return True
            elif sname > oname:
                return False

            return False
    
    def __ne__(self, other):
        return not self == other

    def __le__(self, other):
        return self == other or self < other

    def __ge__(self, other):
        return not self < other

    def __gt__(self, other):
        return not self <= other

    #
    # Connected components
    #

    def is_connected(self):
        r"""
        Test if the strata is connected.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: AbelianStratum([2]).is_connected()
            True
            sage: AbelianStratum([2,2]).is_connected()
            False
            sage: QuadraticStratum([-1,-1,-1,-1]).is_connected()
            True
            sage: QuadraticStratum([12]).is_connected()
            False
        """
        return len(self._cc) <= 1

    def permutation_representative(self, *args, **kwds):
        r"""
        Return a permutation of interval exchanges associated to this stratum.

        EXAMPLES::

            sage: from surface_dynamics import *

        Examples from Abelian differentials::

            sage: a = AbelianStratum([3,2,1,0,0])
            sage: p = a.permutation_representative()
            sage: p.stratum()
            H_4(3, 2, 1, 0^2)
            sage: a = AbelianStratum([2, 2, 2])
            sage: p = a.permutation_representative()
            sage: p.stratum()
            H_4(2^3)

        Examples from quadratic differentials::

            sage: a = QuadraticStratum([6,-1,-1])
            sage: p = a.permutation_representative()
            sage: p.stratum()
            Q_2(6, -1^2)
            sage: a = QuadraticStratum([-1,-1,-1,-1,0,0])
            sage: p = a.permutation_representative()
            sage: p.stratum()
            Q_0(0^2, -1^4)
        """
        return self.one_component().permutation_representative(*args,**kwds)

    def is_empty(self):
        r"""
        Return True if the stratum is empty

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: AbelianStratum(2).is_empty()
            False
            sage: QuadraticStratum(1,-1).is_empty()
            True
        """
        return len(self._cc) == 0

    def number_of_components(self):
        r"""
        Returns the number of connected components of self

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: AbelianStratum(2).number_of_components()
            1
            sage: AbelianStratum(4).number_of_components()
            2
            sage: AbelianStratum(3,3).number_of_components()
            2
        """
        return len(self._cc)

    def one_component(self):
        r"""
        Returns a connected component of this stratum.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: AbelianStratum(2).one_component()
            H_2(2)^hyp
        """
        if self.components():
            return self.components()[-1]
        from sage.categories.sets_cat import EmptySetError
        raise EmptySetError("The stratum is empty")

    def unique_component(self):
        r"""
        Returns the unique component of self or raise a ValueError.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: a = AbelianStratum(1,1); a
            H_2(1^2)
            sage: a.unique_component()
            H_2(1^2)^hyp

            sage: a = AbelianStratum(3,2,1); a
            H_4(3, 2, 1)
            sage: a.unique_component()
            H_4(3, 2, 1)^c

            sage: QuadraticStratum({1:1, -1:5}).unique_component()
            Q_0(1, -1^5)^c
            sage: QuadraticStratum(3,2,-1).unique_component()
            Q_2(3, 2, -1)^nonhyp

            sage: QuadraticStratum(12).unique_component()
            Traceback (most recent call last):
            ...
            ValueError: several components for this stratum
        """
        if len(self._cc) != 1:
            raise ValueError("several components for this stratum")
        return self._cc[0](self)

    def random_component(self):
        r"""
        Returns a random connected component of this stratum.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: Q = QuadraticStratum(6,6)
            sage: Q.random_component()
            Q_4(6^2)^hyp
            sage: Q.random_component()
            Q_4(6^2)^reg
        """
        if self.components():
            from sage.misc.prandom import choice
            return choice(self.components())
        from sage.categories.sets_cat import EmptySetError
        raise EmptySetError("The stratum is empty")

    def components(self):
        """
        Lists the connected components of the Stratum.

        OUTPUT:

        list -- a list of connected components of stratum

        EXAMPLES::

            sage: from surface_dynamics import *

        Some abelian strata::

            sage: AbelianStratum(0).components()
            [H_1(0)^hyp]
            sage: AbelianStratum(2).components()
            [H_2(2)^hyp]
            sage: AbelianStratum(4).components()
            [H_3(4)^hyp, H_3(4)^odd]
            sage: AbelianStratum(2,2).components()
            [H_3(2^2)^hyp, H_3(2^2)^odd]
            sage: AbelianStratum(1,1,1,1).components()
            [H_3(1^4)^c]

        Some quadratic strata::

            sage: QuadraticStratum(12).components()
            [Q_4(12)^reg, Q_4(12)^irr]
            sage: QuadraticStratum(6,-1,-1).components()
            [Q_2(6, -1^2)^hyp, Q_2(6, -1^2)^nonhyp]
        """
        return list(map(lambda x: x(self), self._cc))

class StratumComponent(SageObject):
    r"""
    Generic class for connected component of a stratum of flat surfaces.

    Assumes there are implemented

    - a method .permutation_representative()

    There may be

    - an attribute ._name

    - an attribute ._latex_name

    """
    _name = ''
    _latex_name = ''

    def __init__(self, stratum):
        r"""
        TEST::

            sage: from surface_dynamics import *

            sage: a = AbelianStratum(4,4).one_component()
            sage: a == loads(dumps(a))
            True
            sage: q = QuadraticStratum(5,5,-1,-1).one_component()
            sage: q == loads(dumps(q))
            True
        """
        self._stratum = stratum

    def __reduce__(self):
        r"""
        Reduce method for pickling

        TESTS::

            sage: from surface_dynamics import *

        Tests for Abelian strata::

            sage: a = AbelianStratum(2,2)
            sage: all(loads(dumps(cc)) == cc for cc in a.components())
            True
            sage: a = AbelianStratum(3,3)
            sage: all(loads(dumps(cc)) == cc for cc in a.components())
            True
            sage: a = AbelianStratum(6)
            sage: all(loads(dumps(cc)) == cc for cc in a.components())
            True
            sage: a = AbelianStratum(1,1,1,1)
            sage: all(loads(dumps(cc)) == cc for cc in a.components())
            True

        Tests for quadratic strata::

            sage: q = QuadraticStratum(-1,-1,-1,-1)
            sage: all(loads(dumps(cc)) == cc for cc in q.components())
            True
            sage: q = QuadraticStratum(12)
            sage: all(loads(dumps(cc)) == cc for cc in q.components())
            True
        """
        return (self.__class__, (self._stratum,))

    def _repr_(self):
        r"""
        String representation

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: a_hyp = AbelianStratum(4).hyperelliptic_component()
            sage: a_hyp._repr_()
            'H_3(4)^hyp'
            sage: a_odd = AbelianStratum(4).odd_component()
            sage: a_odd._repr_()
            'H_3(4)^odd'
        """
        return str(self._stratum) + "^" + self._name

    def __hash__(self):
        r"""
        TESTS::

            sage: from surface_dynamics import *

            sage: A4hyp = AbelianStratum(4).hyperelliptic_component()
            sage: A4odd = AbelianStratum(4).odd_component()
            sage: hash(A4hyp) != hash(A4odd)
            True
            sage: A22hyp = AbelianStratum(2,2).hyperelliptic_component()
            sage: hash(A22hyp) != hash(A4hyp)
            True
        """
        return hash(self._stratum) ^ hash(self._name)

    def stratum(self):
        r"""
        Return the stratum associated to self

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: a = AbelianStratum(4,4)
            sage: all([c.stratum() == a for c in a.components()])
            True
        """
        return self._stratum

    def genus(self):
        r"""
        Return genus of the corresponding stratum

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: a = AbelianStratum(4,4)
            sage: a.one_component().genus()
            5
        """
        return self._stratum.genus()

    def dimension(self):
        r"""
        Return the (complex) dimension of this GL(2,R)-invariant orbifold.

        EXAMPLES::

            sage: from surface_dynamics import AbelianStratum, QuadraticStratum

            sage: AbelianStratum(4).odd_component().dimension()
            6
            sage: QuadraticStratum(12).regular_component().dimension()
            7
        """
        return self._stratum.dimension()

    def rank(self):
        r"""
        Return the rank of this GL(2,R)-invariant orbifold.

        EXAMPLES::

            sage: from surface_dynamics import AbelianStratum, QuadraticStratum

            sage: AbelianStratum(4).odd_component().rank()
            3
            sage: QuadraticStratum(12).regular_component().rank()
            3
        """
        return self._stratum.rank()

    def __eq__(self,other):
        r"""
        Equality test

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: c_hyp = AbelianStratum(6).hyperelliptic_component()
            sage: c_odd = AbelianStratum(6).odd_component()
            sage: c_hyp == c_hyp
            True
            sage: c_hyp == c_odd
            False
        """
        if not isinstance(self, StratumComponent) or not isinstance(other, StratumComponent):
            return NotImplemented

        return type(self) == type(other) and self._stratum == other._stratum

    def __lt__(self, other):
        r"""
        Comparison

        TESTS::

            sage: from surface_dynamics import *

            sage: a1 = AbelianStratum(1,1,1,1)
            sage: c1 = a1.components()[0]
            sage: a2 = AbelianStratum(3,1)
            sage: c2 = a2.components()[0]
            sage: c1 == c1
            True
            sage: c1 == c2
            False
            sage: a1 = AbelianStratum(1,1,1,1)
            sage: c1 = a1.components()[0]
            sage: a2 = AbelianStratum(2, 2)
            sage: c2_hyp, c2_odd = a2.components()
            sage: c1 != c1
            False
            sage: c1 != c2_hyp
            True
            sage: c2_hyp != c2_odd
            True
        """
        if not isinstance(self, StratumComponent) or not isinstance(other, StratumComponent):
            return NotImplemented

        if self._stratum < other._stratum:
            return True
        elif self._stratum > other._stratum:
            return False

        if type(self) < type(other):
            return True
        elif type(self) > type(other):
            return False

#
# Strata (family of strata)
#

class Strata(Parent):
    r"""
    Strata of Abelian or Quadratic differentials.
    """
    pass



#*****************************************************************************
#       Copyright (C) 2019 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

from sage.cpython.string import bytes_to_str, str_to_bytes

def iscc_rg_string(r):
    r"""
    EXAMPLES::

        sage: from surface_dynamics import *
        sage: from surface_dynamics.misc.iscc import iscc_rg_string
        sage: r = RibbonGraph(faces='(0,2,4,6,7)(8,9,5,3,1)', edges='(0,1)(2,3)(4,5)(6,7)(8,9)')
        sage: iscc_rg_string(r)
        '[b0, b1] -> { [l0, l1, l2, l3, l4, l5, l6, l7, l8, l9] : l0 >= 0 and l1 >= 0 and l2 >= 0 and l3 >= 0 and l4 >= 0 and l5 >= 0 and l6 >= 0 and l7 >= 0 and l8 >= 0 and l9 >= 0 and l0 = l1 and l2 = l3 and l4 = l5 and l6 = l7 and l8 = l9 and b0 = l0 + l2 + l4 + l6 + l7 and b1 = l1 + l8 + l9 + l5 + l3 }'
    """
    boundary_vars = ', '.join('b%d'%i for i in range(r.num_faces()))
    lengths_vars = ', '.join('l%d'%i for i in r.darts())
    
    ieqs = ' and '.join('l%d >= 0' % i for i in r.darts())
    edges = ' and '.join('l%d = l%d' %(i,j) for i,j in r.edges())
    
    eqns = []
    for i,f in enumerate(r.faces()):
        eqns.append('b%d = %s' % (i, ' + '.join('l%d' %j for j in f)))
    eqns = ' and '.join(eqns)
    
    return "[{boundary_vars}] -> {{ [{lengths_vars}] : {ieqs} and {edges} and {eqns} }}".format(
               boundary_vars=boundary_vars,
               lengths_vars=lengths_vars,
               edges=edges,
               ieqs=ieqs,
               eqns=eqns)

def iscc_cd_string(cd):
    r"""
    Given a cylinder diagram produces a parametrized polytopes (by the widths
    of cylinders)

    EXAMPLES::

        sage: from surface_dynamics.misc.iscc import iscc_cd_string
        sage: from surface_dynamics import CylinderDiagram
        sage: cd = CylinderDiagram('(0,1)-(0,2) (2)-(1)')
        sage: iscc_cd_string(cd)
        '[w0, w1] -> { [l0, l1, l2] : l0 >= 0 and l1 >= 0 and l2 >= 0 and w0 = l0 + l1 = l0 + l2 and w1 = l2 = l1 }'
    """
    widths_vars = ', '.join('w%d'%i for i in range(cd.ncyls()))
    lengths_vars = ', '.join('l%d'%i for i in range(cd.num_edges()))
    
    ieqs = []
    for i in range(cd.num_edges()):
        ieqs.append('l%d >= 0' % i)
    ieqs = ' and '.join(ieqs)
    
    eqns = []
    for i,(top,bot) in enumerate(cd.cylinders()):
        eqns.append('w%d = %s = %s' % (i,
                        ' + '.join('l%d' %j for j in top),
                        ' + '.join('l%d' %j for j in bot)))
    eqns = ' and '.join(eqns)
    
    return "[{widths_vars}] -> {{ [{lengths_vars}] : {ieqs} and {eqns} }}".format(
               widths_vars = widths_vars,
               lengths_vars=lengths_vars,
               ieqs=ieqs,
               eqns=eqns)

def iscc_card(arg):
    r"""
    EXAMPLES::

        sage: from surface_dynamics import *
        sage: from surface_dynamics.misc.iscc import iscc_card
        sage: cd = CylinderDiagram('(0,1)-(0,2) (2)-(1)')
        sage: iscc_card(cd)  # optional - barvinok
        '[w0, w1] -> { 1 : 0 <= w1 <= w0 }'

        sage: r = RibbonGraph(faces='(0,2,4,6,7)(8,9,5,3,1)', edges='(0,1)(2,3)(4,5)(6,7)(8,9)')
        sage: iscc_card(r)   # optional - barvinok
        '[b0, b1] -> { (((1 - 1/8 * b0) + 17/12 * b1 + 5/8 * b1^2 + 1/12 * b1^3) + 1/4 * floor((2b0)/4)) : (b0 + b1) mod 2 = 0 and 0 <= b1 <= -2 + b0; ((1 + 31/24 * b0 + 5/8 * b0^2 + 1/12 * b0^3) + 1/4 * floor((2b0)/4)) : (b0 + b1) mod 2 = 0 and b0 >= 0 and b1 >= b0 }'
    """
    from surface_dynamics.flat_surfaces.separatrix_diagram import CylinderDiagram
    from surface_dynamics.flat_surfaces.homology import RibbonGraph

    if isinstance(arg, CylinderDiagram):
        poly = iscc_cd_string(arg)
    elif isinstance(arg, RibbonGraph):
        poly = iscc_rg_string(arg)

    from subprocess import Popen, PIPE
    
    cmd = 'P := {};\n'.format(poly)
    cmd += 'q := card P;\n'
    cmd += 'q;'
    
    proc = Popen(['iscc'], stdin=PIPE, stdout=PIPE, stderr=PIPE)
    ans, err = proc.communicate(str_to_bytes(cmd))
    
    return bytes_to_str(ans).strip()

def parse_iscc_result(ans, d):
    R = PolynomialRing(QQ, ['w%d'%i for i in range(d)], d)
    gens = {'w%d'%i: R.gen(i) for i in range(d)}
    prefix = '[' + ', '.join('w%d'%i for i in range(d)) + ']'
    prefix += ' -> { '
    suffix = ' }\n'
    if not ans.startswith(prefix):
        raise ValueError('bad prefix')
    if not ans.endswith(suffix):
        raise ValueError('bad suffix')
    pieces = ans[len(prefix):len(ans)-len(suffix)].split(';')
    res = []
    for piece in pieces:
        if piece.count(':') != 1:
            raise ValueError('parsing error!')
        i = piece.find(':')
        poly = R(sage_eval(piece[:i], locals=gens))
        cond = piece[i+1:]
        res.append((cond,poly))
    return res

def higher_order_terms(p):
    R = p.parent()
    p_trunc = R.zero()
    deg = p.degree()
    for coeff, exp in zip(p.coefficients(), p.exponents()):
        if sum(exp) == deg:
            p_trunc += R({exp:coeff})
    return p_trunc
